/* Disable overriding memory allocators. */
/* #undef DXC_DISABLE_ALLOCATOR_OVERRIDES */
/* Generate a trap if an hlsl::Exception is thrown during code generation */
/* #undef DXC_CODEGEN_EXCEPTIONS_TRAP */
